// This may look like C code, but it is really -*- C++ -*-
/*
   wordlattice - a well known word game
   Copyright (C) 1997  Wolfgang Dautermann

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

 */

#define VERSION "wordlattice: Version 1.0 \n"

#define HILFE "Usage: wordlattice [options] word ... \n \n" \
              "-h --help                  Output this help \n" \
              "-v --version               Version \n" \
              "-oxxx --output=xxx         Output file xxx \n" \
              "-sxxx --solution=xxx       Solution file  xxx \n" \
              "-cxxx --columns=xxx        Number of columns \n" \
              "-lxxx --lines=xxx          Number of lines \n"


#define OPTIONFEHLER "Try wordlattice --help' for more information \n"

#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <getopt.h>

// Number of trials to insert a word
#define DEFAULT_TRIALS 100

class wordlattice {
	private:
	int columns;
	int lines;
	int trials;
	char *char_field;
	int *used_field;
	void fillrandom();
	int isfree(int, int, int, int, char *);
	 public:
	 wordlattice(int, int);	// constructor

	~wordlattice();		// destructor

	void print(FILE *);
	void printsolution(FILE *);
	int insertword(char *);
};



class datenklasse {
	public:
	int columns;
	int lines;
	char *o_filename;
	char *s_filename;
	char **words;		// array of words to insert

	int numberofwords;
	int execution;		// 1: immidiate exit or 0: execution of the program

	 datenklasse(int);	// constructor

	~datenklasse();		// destructor

	void optionen(int, char **);	// read the options and modify the defaultdata

};



wordlattice::wordlattice(int c, int l)
{
	columns = c;
	lines = l;
	trials = DEFAULT_TRIALS;
	srand(time(0));		// initialize the random generator

	char_field = (char *) malloc(columns * lines * sizeof(char));
	used_field = (int *) malloc(columns * lines * sizeof(int));
	fillrandom();
}

wordlattice::~wordlattice()
{
	free(char_field);
	free(used_field);
}

void wordlattice::fillrandom()
{
	int x, y;
	for (x = 0; x < columns; x++) {
		for (y = 0; y < lines; y++) {
			char c;
			c = rand() % 26 + 'A';
			char_field[y * columns + x] = c;
			used_field[y * columns + x] = 0;	// no field is used 

		}
	}
}

// return 1, if insertion is not allowed
int wordlattice::isfree(int x, int y, int xs, int ys, char *w)
{
	int len, i, xx, yy;

	len = strlen(w);

	// either xs or ys must be != 0 
	if ((xs == 0) && (ys == 0))
		return 1;

	// word must fit in the lattice
	if (x + xs * (len - 1) < 0)
		return 1;
	if (y + ys * (len - 1) < 0)
		return 1;
	if (x + xs * (len - 1) >= columns)
		return 1;
	if (y + ys * (len - 1) >= lines)
		return 1;

	// word fits in the lattice
	// no currently used letter may change
	xx = x;
	yy = y;
	for (i = 0; i < len; i++) {
		if ((used_field[yy * columns + xx] != 0) &&
		    (char_field[yy * columns + xx] != w[i]))
			return 1;
		xx = xx + xs;
		yy = yy + ys;
	};
	// all tests passed - return 0 ;
	return 0;
}

int wordlattice::insertword(char *w)
{
	int len, i, x, y, xs, ys, xx, yy, t;

	len = strlen(w);
	for (t = 0; t < trials; t++) {
		x = rand() % columns;
		y = rand() % lines;
		xs = (rand() % 3) - 1;	// -1, 0 or 1

		ys = (rand() % 3) - 1;	// -1, 0 or 1

		if (isfree(x, y, xs, ys, w) == 0) {
			xx = x;
			yy = y;
			for (i = 0; i < len; i++) {
				char_field[yy * columns + xx] = w[i];
				used_field[yy * columns + xx]++;
				xx = xx + xs;
				yy = yy + ys;
			};
			return 0;
		};		// end if

	};
	// insertion was not possible
	return -1;
}

void wordlattice::print(FILE * f)
{
	int x, y;
	for (y = 0; y < lines; y++) {
		for (x = 0; x < columns; x++) {
			fprintf(f, "%c", char_field[y * columns + x]);
		};
		fprintf(f, "\n");
	}
}

void wordlattice::printsolution(FILE * f)
{
	int x, y;
	for (y = 0; y < lines; y++) {
		for (x = 0; x < columns; x++) {
			if (used_field[y * columns + x] != 0) {
				fprintf(f, "%c", char_field[y * columns + x]);
			} else {
				fprintf(f, ".");
			};
		};
		fprintf(f, "\n");
	}
}


datenklasse::datenklasse(int number)
{
	columns = 20;
	lines = 10;
	o_filename = "";
	s_filename = "";
	words = (char **) malloc(number * sizeof(char *));
	numberofwords = 0;
	execution = 0;		// 1: immidiate exit or 0: execution of the program

}


datenklasse::~datenklasse()
{
	free(words);
}

void datenklasse::optionen(int argc, char **argv)
{				/* I used the example in the manpage for getopt(3) */
	int c;
	int digit_optind = 0;

	while (1) {
		int this_option_optind = optind ? optind : 1;
		int option_index = 0;
		static struct option long_options[] =
		{
			{"version", no_argument, 0, 'v'},
			{"help", no_argument, 0, 'h'},
			{"output", required_argument, 0, 'o'},
			{"solution", required_argument, 0, 's'},
			{"columns", required_argument, 0, 'c'},
			{"lines", required_argument, 0, 'l'},
			{0, 0, 0, 0}
		};

		c = getopt_long(argc, argv, "vho:s:c:l:",
				long_options, &option_index);
		if (c == -1)
			break;

		switch (c) {
		case 0:	/* impossible: For every --opt there is a -opt */
			break;

		case 'v':	/* Versionsinformation */
			printf(VERSION);
			execution = 1;	// exit immediately

			break;

		case 'h':	/* Hilfe */
			printf(HILFE);
			execution = 1;	// exit immediately

			break;

		case 'o':	/* output file */
			o_filename = optarg;
			break;

		case 's':	/* solution file */
			s_filename = optarg;
			break;

		case 'c':	/* columns */
			columns = atoi(optarg);
			break;

		case 'l':	/* lines */
			lines = atoi(optarg);
			break;

		case '?':
			printf(OPTIONFEHLER);
			break;

		default:
			printf("?? getopt returned character code 0%o ??\n", c);
		}
	}
	while (optind < argc) {	/* non option ARGV-elements: words for insertion */
		words[numberofwords] = argv[optind];
		optind++;
		numberofwords++;
	};
}				/* Ende optionenhandling */


int main(int argc, char **argv)
{
	FILE *outfile;
	FILE *solutionfile;

	datenklasse pdat(argc);	// Defaultwerte setzen

	pdat.optionen(argc, argv);	// Defaultwerte durch Optionen aendern?

	if (pdat.execution == 1)
		return 0;	// version or help information 

	wordlattice lattice(pdat.columns, pdat.lines);

	while (pdat.numberofwords > 0) {
		int success;
		pdat.numberofwords--;
		success = lattice.insertword(pdat.words[pdat.numberofwords]);
		if (success == -1) {
			fprintf(stderr, "%s: Insertion not successfull\n",
				pdat.words[pdat.numberofwords]);
		};
	};

	if (strlen(pdat.o_filename) > 0)
		outfile = fopen(pdat.o_filename, "w");
	else
		outfile = stdout;
	lattice.print(outfile);
	if (strlen(pdat.o_filename) > 0)
		fclose(outfile);

	if (strlen(pdat.s_filename) > 0)
		solutionfile = fopen(pdat.s_filename, "w");
	else
		solutionfile = stdout;
	lattice.printsolution(solutionfile);
	if (strlen(pdat.s_filename) > 0)
		fclose(solutionfile);
}
